package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.Likes;
import at.ac.uibk.gitsearch.repository.LikesRepository;
import at.ac.uibk.gitsearch.repository.search.LikesSearchRepository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import static org.elasticsearch.index.query.QueryBuilders.*;

/**
 * Service Implementation for managing {@link Likes}.
 */
@Service
@Transactional
public class LikesService {

    private final Logger log = LoggerFactory.getLogger(LikesService.class);

    private final LikesRepository likesRepository;

    private final LikesSearchRepository likesSearchRepository;

    public LikesService(LikesRepository likesRepository, LikesSearchRepository likesSearchRepository) {
        this.likesRepository = likesRepository;
        this.likesSearchRepository = likesSearchRepository;
    }

    /**
     * Save a likes.
     *
     * @param likes the entity to save.
     * @return the persisted entity.
     */
    public Likes save(Likes likes) {
        log.debug("Request to save Likes : {}", likes);
        Likes result = likesRepository.save(likes);
        likesSearchRepository.save(result);
        return result;
    }

    /**
     * Get all the likes.
     *
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public List<Likes> findAll() {
        log.debug("Request to get all Likes");
        return likesRepository.findAll();
    }


    /**
     * Get one likes by id.
     *
     * @param id the id of the entity.
     * @return the entity.
     */
    @Transactional(readOnly = true)
    public Optional<Likes> findOne(Long id) {
        log.debug("Request to get Likes : {}", id);
        return likesRepository.findById(id);
    }

    /**
     * Delete the likes by id.
     *
     * @param id the id of the entity.
     */
    public void delete(Long id) {
        log.debug("Request to delete Likes : {}", id);
        likesRepository.deleteById(id);
        likesSearchRepository.deleteById(id);
    }

    public void deleteByUserIDandExerciseID(Integer userID, String exerciseID) {
        Likes like = findLikesByUserIDandExerciseID(userID, exerciseID);
        if (like.getId() != null){
            delete(like.getId());
        }
    }

    /**
     * Search for the likes corresponding to the query.
     *
     * @param query the query of the search.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public List<Likes> search(String query) {
        log.debug("Request to search Likes for query {}", query);
        return StreamSupport
            .stream(likesSearchRepository.search(queryStringQuery(query)).spliterator(), false)
        .collect(Collectors.toList());
    }

    @Transactional(readOnly = true)
    public Likes findLikesByUserIDandExerciseID(Integer userID, String exerciseID) {
        log.debug("Request to get Likes by User id and project id : {} {}", userID, exerciseID);
        return likesRepository.findLikesByUserIDandExerciseID(userID, exerciseID);
    }

    @Transactional(readOnly = true)
    public Integer findNumberOfLikesByExerciseID(String exerciseID) {
        log.debug("Request to get number of Likes by project id : {}", exerciseID);
        return likesRepository.findNumberOfLikesByExerciseID(exerciseID);
    }
}
