package at.ac.uibk.gitsearch.web.rest;

import static org.assertj.core.api.Assertions.assertThat;
import static org.elasticsearch.index.query.QueryBuilders.queryStringQuery;
import static org.hamcrest.Matchers.hasItem;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.put;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import javax.persistence.EntityManager;

import org.elasticsearch.node.NodeValidationException;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.transaction.annotation.Transactional;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.domain.Statistics;
import at.ac.uibk.gitsearch.repository.StatisticsRepository;
import at.ac.uibk.gitsearch.repository.search.StatisticsSearchRepository;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestServerConfiguration;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.StatisticsService;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.service.mapper.StatisticsMapper;

/**
 * Integration tests for the {@link StatisticsResource} REST controller.
 */
@SpringBootTest(classes = GitsearchApp.class)
@ExtendWith(MockitoExtension.class)
@AutoConfigureMockMvc
@WithMockUser
public class StatisticsResourceIT {

    private static final Integer DEFAULT_VIEWS = 1;
    private static final Integer UPDATED_VIEWS = 2;

    private static final Integer DEFAULT_DOWNLOADS = 1;
    private static final Integer UPDATED_DOWNLOADS = 2;

    private static final String DEFAULT_EXERCISE_ID = "1";
    private static final String UPDATED_EXERCISE_ID = "2";
    
    final String publicGitLabProjectId = "3";

    @Autowired
    private StatisticsRepository statisticsRepository;

    @Autowired
    private StatisticsMapper statisticsMapper;

    @Autowired
    private StatisticsService statisticsService;

    /**
     * This repository is mocked in the at.ac.uibk.gitsearch.repository.search test
     * package.
     *
     * @see at.ac.uibk.gitsearch.repository.search.StatisticsSearchRepositoryMockConfiguration
     */
    @Autowired
    private StatisticsSearchRepository mockStatisticsSearchRepository;

    @Autowired
    private EntityManager em;

    @Autowired
    private MockMvc restStatisticsMockMvc;

    private Statistics statistics;


    @BeforeAll
	public static void setUpESServer() throws IOException, NodeValidationException {
		ElasticSearchTestServerConfiguration.getTestInstance().startTestNode();
	}

    /**
     * Create an entity for this test.
     *
     * This is a static method, as tests for other entities might also need it, if
     * they test an entity which requires the current entity.
     */
    public static Statistics createEntity(EntityManager em) {
        Statistics statistics = new Statistics().views(DEFAULT_VIEWS).downloads(DEFAULT_DOWNLOADS)
                .exerciseID(DEFAULT_EXERCISE_ID);
        return statistics;
    }

    /**
     * Create an updated entity for this test.
     *
     * This is a static method, as tests for other entities might also need it, if
     * they test an entity which requires the current entity.
     */
    public static Statistics createUpdatedEntity(EntityManager em) {
        Statistics statistics = new Statistics().views(UPDATED_VIEWS).downloads(UPDATED_DOWNLOADS)
                .exerciseID(UPDATED_EXERCISE_ID);
        return statistics;
    }

    @BeforeEach
    public void initTest() {
        statistics = createEntity(em);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void createStatistics() throws Exception {
        int databaseSizeBeforeCreate = statisticsRepository.findAll().size();
        // Create the Statistics
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(statistics);
        restStatisticsMockMvc.perform(post("/api/statistics").with(csrf().asHeader())
                .contentType(MediaType.APPLICATION_JSON).content(TestUtil.convertObjectToJsonBytes(statisticsDTO)))
                .andExpect(status().isCreated());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeCreate + 1);
        Statistics testStatistics = statisticsList.get(statisticsList.size() - 1);
        assertThat(testStatistics.getViews()).isEqualTo(DEFAULT_VIEWS);
        assertThat(testStatistics.getDownloads()).isEqualTo(DEFAULT_DOWNLOADS);
        assertThat(testStatistics.getExerciseID()).isEqualTo(DEFAULT_EXERCISE_ID);

        // Validate the Statistics in Elasticsearch
        verify(mockStatisticsSearchRepository, times(1)).save(testStatistics);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void createStatisticsWithExistingId() throws Exception {
        int databaseSizeBeforeCreate = statisticsRepository.findAll().size();

        // Create the Statistics with an existing ID
        statistics.setId(1L);
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(statistics);

        // An entity with an existing ID cannot be created, so this API call must fail
        restStatisticsMockMvc.perform(post("/api/statistics").with(csrf().asHeader())
                .contentType(MediaType.APPLICATION_JSON).content(TestUtil.convertObjectToJsonBytes(statisticsDTO)))
                .andExpect(status().isBadRequest());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeCreate);

        // Validate the Statistics in Elasticsearch
        verify(mockStatisticsSearchRepository, times(0)).save(statistics);
    }


    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void getAllStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        // Get all the statisticsList
        restStatisticsMockMvc.perform(get("/api/statistics?sort=id,desc").with(csrf().asHeader())).andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
                .andExpect(jsonPath("$.[*].id").value(hasItem(statistics.getId().intValue())))
                .andExpect(jsonPath("$.[*].views").value(hasItem(DEFAULT_VIEWS)))
                .andExpect(jsonPath("$.[*].downloads").value(hasItem((int)(long)DEFAULT_DOWNLOADS)))
                .andExpect(jsonPath("$.[*].exerciseID").value(hasItem(DEFAULT_EXERCISE_ID)));
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void getStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        // Get the statistics
        restStatisticsMockMvc.perform(get("/api/statistics/{id}", statistics.getId()).with(csrf().asHeader())).andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
                .andExpect(jsonPath("$.id").value(statistics.getId().intValue()))
                .andExpect(jsonPath("$.views").value(DEFAULT_VIEWS))
                .andExpect(jsonPath("$.downloads").value(DEFAULT_DOWNLOADS))
                .andExpect(jsonPath("$.exerciseID").value(DEFAULT_EXERCISE_ID));
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void testAchievementCalculation() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        // Get the statistics
        restStatisticsMockMvc.perform(get("/user/achievements").with(csrf().asHeader())).andExpect(status().isOk());
    
    }


    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void getNonExistingStatistics() throws Exception {
        // Get the statistics
        restStatisticsMockMvc.perform(get("/api/statistics/{id}", Long.MAX_VALUE).with(csrf().asHeader())).andExpect(status().isNotFound());
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void updateStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        int databaseSizeBeforeUpdate = statisticsRepository.findAll().size();

        // Update the statistics
        Statistics updatedStatistics = statisticsRepository.findById(statistics.getId()).get();
        // Disconnect from session so that the updates on updatedStatistics are not
        // directly saved in db
        em.detach(updatedStatistics);
        updatedStatistics.views(UPDATED_VIEWS).downloads(UPDATED_DOWNLOADS).exerciseID(UPDATED_EXERCISE_ID);
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(updatedStatistics);

        restStatisticsMockMvc.perform(put("/api/statistics").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON)
                .content(TestUtil.convertObjectToJsonBytes(statisticsDTO))).andExpect(status().isOk());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeUpdate);
        Statistics testStatistics = statisticsList.get(statisticsList.size() - 1);
        assertThat(testStatistics.getViews()).isEqualTo(UPDATED_VIEWS);
        assertThat(testStatistics.getDownloads()).isEqualTo(UPDATED_DOWNLOADS);
        assertThat(testStatistics.getExerciseID()).isEqualTo(UPDATED_EXERCISE_ID);

        // Validate the Statistics in Elasticsearch
        verify(mockStatisticsSearchRepository, times(1)).save(testStatistics);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void updateNonExistingStatistics() throws Exception {
        int databaseSizeBeforeUpdate = statisticsRepository.findAll().size();

        // Create the Statistics
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(statistics);

        // If the entity doesn't have an ID, it will throw BadRequestAlertException
        restStatisticsMockMvc.perform(put("/api/statistics").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON)
                .content(TestUtil.convertObjectToJsonBytes(statisticsDTO))).andExpect(status().isBadRequest());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeUpdate);

        // Validate the Statistics in Elasticsearch
        verify(mockStatisticsSearchRepository, times(0)).save(statistics);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void deleteStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        int databaseSizeBeforeDelete = statisticsRepository.findAll().size();

        // Delete the statistics
        restStatisticsMockMvc
                .perform(delete("/api/statistics/{id}", statistics.getId()).with(csrf().asHeader()).accept(MediaType.APPLICATION_JSON))
                .andExpect(status().isNoContent());

        // Validate the database contains one less item
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeDelete - 1);

        // Validate the Statistics in Elasticsearch
        verify(mockStatisticsSearchRepository, times(1)).deleteById(statistics.getId());
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void searchStatistics() throws Exception {
        // Configure the mock search repository
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);
        when(mockStatisticsSearchRepository.search(queryStringQuery("id:" + statistics.getId()), PageRequest.of(0, 20)))
                .thenReturn(new PageImpl<>(Collections.singletonList(statistics), PageRequest.of(0, 1), 1));

        // Search the statistics
        restStatisticsMockMvc.perform(get("/api/_search/statistics?query=id:" + statistics.getId()).with(csrf().asHeader()))
                .andExpect(status().isOk()).andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
                .andExpect(jsonPath("$.[*].id").value(hasItem(statistics.getId().intValue())))
                .andExpect(jsonPath("$.[*].views").value(hasItem(DEFAULT_VIEWS)))
                .andExpect(jsonPath("$.[*].downloads").value(hasItem(DEFAULT_DOWNLOADS)))
                .andExpect(jsonPath("$.[*].exerciseID").value(hasItem(DEFAULT_EXERCISE_ID)));
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void getStatisticsByExerciseId() throws Exception {


        // Get the statistics and see if it coincides with the mock stats

        statisticsRepository.deleteAll();

        //After the following call a new statistics entity for the exerciseID 143 should have been created because there exists a resource with this id in elastic search from stefan podlipnig
        restStatisticsMockMvc.perform(get("/api/statistics/exercise/{id}", "143").with(csrf().asHeader())).andExpect(status().isOk());
        assertEquals(statisticsService.findOneByExerciseID("143").get().getExerciseID(), "143");

        //Do the same call again to see if the statisticsService understood that the entity already exists and it only needs to increase the views
        restStatisticsMockMvc.perform(get("/api/statistics/exercise/{id}", "143").with(csrf().asHeader())).andExpect(status().isOk());
        assertEquals(statisticsService.findOneByExerciseID("143").get().getViews(), 2);


        //Test the search for an id which does not exist
        assertEquals(Optional.empty(),statisticsService.findOneByExerciseID("notThere"));

    }


    @ParameterizedTest
    @ValueSource(strings = {"3:with/path"})
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    public void createStatisticsByExerciseIdNotFoundWithES(String exerciseId) throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        restStatisticsMockMvc.perform(get("/api/statistics/exercise/{id}", exerciseId).with(csrf().asHeader())).andExpect(status().isNotFound());
        
        
        // Get the statistics and see if it is found
        assertEquals(Optional.empty(),statisticsService.findOneByExerciseID(exerciseId));
    }




}
